@@ -37,6 +37,7 @@
 #include <unistd.h>
 /* clang-format off */
 #include <xcb/xcb.h>
+#include <xcb/shm.h>
 #include <vulkan/vulkan_core.h>
 #include <vulkan/vulkan_xcb.h>
 /* clang-format on */
@@ -46,12 +47,14 @@
 #include "util/bitscan.h"
 #include "util/macros.h"
 #include "util/u_debug.h"
+#include "util/log.h"
 #include <sys/stat.h>
 #include <sys/types.h>
 
 #include "kopper_interface.h"
 #include "loader.h"
 #include "platform_x11.h"
+#include "drm-uapi/drm_fourcc.h"
 
 #ifdef HAVE_DRI3
 #include "platform_x11_dri3.h"
@@ -150,14 +153,14 @@ swrastGetDrawableInfo(__DRIdrawable *draw, int *x, int *y, int *w, int *h,
 }
 
 static void
-swrastPutImage(__DRIdrawable *draw, int op, int x, int y, int w, int h,
-               char *data, void *loaderPrivate)
+swrastPutImage2(__DRIdrawable *draw, int op, int x, int y, int w, int h,
+                int stride, char *data, void *loaderPrivate)
 {
    struct dri2_egl_surface *dri2_surf = loaderPrivate;
    struct dri2_egl_display *dri2_dpy =
       dri2_egl_display(dri2_surf->base.Resource.Display);
-   size_t hdr_len = sizeof(xcb_put_image_request_t);
    int stride_b = dri2_surf->bytes_per_pixel * w;
+   size_t hdr_len = sizeof(xcb_put_image_request_t);
    size_t size = (hdr_len + stride_b * h) >> 2;
    uint64_t max_req_len = xcb_get_maximum_request_length(dri2_dpy->conn);
 
@@ -174,7 +177,23 @@ swrastPutImage(__DRIdrawable *draw, int op, int x, int y, int w, int h,
       return;
    }
 
-   if (size < max_req_len) {
+   /* clamp to drawable size */
+   if (y + h > dri2_surf->base.Height)
+      h = dri2_surf->base.Height - y;
+
+   /* If stride of pixels to copy is different from the surface stride
+    * then we need to copy lines one by one.
+    */
+   if (stride_b != stride) {
+      for (unsigned i = 0; i < h; i++) {
+         cookie = xcb_put_image(
+            dri2_dpy->conn, XCB_IMAGE_FORMAT_Z_PIXMAP, dri2_surf->drawable, gc, w,
+            1, x, y+i, 0, dri2_surf->depth, stride_b, (uint8_t*)data);
+         xcb_discard_reply(dri2_dpy->conn, cookie.sequence);
+
+         data += stride;
+      }
+   } else if (size < max_req_len) {
       cookie = xcb_put_image(
          dri2_dpy->conn, XCB_IMAGE_FORMAT_Z_PIXMAP, dri2_surf->drawable, gc, w,
          h, x, y, 0, dri2_surf->depth, h * stride_b, (const uint8_t *)data);
@@ -195,11 +214,22 @@ swrastPutImage(__DRIdrawable *draw, int op, int x, int y, int w, int h,
          y_todo -= this_lines;
       }
    }
+   xcb_flush(dri2_dpy->conn);
 }
 
 static void
-swrastGetImage(__DRIdrawable *read, int x, int y, int w, int h, char *data,
-               void *loaderPrivate)
+swrastPutImage(__DRIdrawable *draw, int op, int x, int y, int w, int h,
+               char *data, void *loaderPrivate)
+{
+   struct dri2_egl_surface *dri2_surf = loaderPrivate;
+   int stride_b = dri2_surf->bytes_per_pixel * w;
+   swrastPutImage2(draw, op, x, y, w, h, stride_b, data, loaderPrivate);
+}
+
+static void
+swrastGetImage2(__DRIdrawable * read,
+                int x, int y, int w, int h, int stride,
+                char *data, void *loaderPrivate)
 {
    struct dri2_egl_surface *dri2_surf = loaderPrivate;
    struct dri2_egl_display *dri2_dpy =
@@ -221,11 +251,120 @@ swrastGetImage(__DRIdrawable *read, int x, int y, int w, int h, char *data,
    } else {
       uint32_t bytes = xcb_get_image_data_length(reply);
       uint8_t *idata = xcb_get_image_data(reply);
-      memcpy(data, idata, bytes);
+      int stride_b = w * dri2_surf->bytes_per_pixel;
+      /* Only copy line by line if we have a different stride */
+      if (stride != stride_b) {
+         for (int i = 0; i < h; i++) {
+            memcpy(data, idata, stride_b);
+            data += stride;
+            idata += stride_b;
+         }
+      } else {
+         memcpy(data, idata, bytes);
+      }
    }
    free(reply);
 }
 
+static void
+swrastGetImage(__DRIdrawable *read, int x, int y, int w, int h, char *data,
+               void *loaderPrivate)
+{
+   struct dri2_egl_surface *dri2_surf = loaderPrivate;
+   int stride_b = w * dri2_surf->bytes_per_pixel;
+   swrastGetImage2(read, x, y, w, h, stride_b, data, loaderPrivate);
+}
+
+static void
+swrastPutImageShm(__DRIdrawable * draw, int op,
+                  int x, int y, int w, int h, int stride,
+                  int shmid, char *shmaddr, unsigned offset,
+                  void *loaderPrivate)
+{
+   struct dri2_egl_surface *dri2_surf = loaderPrivate;
+   struct dri2_egl_display *dri2_dpy =
+      dri2_egl_display(dri2_surf->base.Resource.Display);
+   xcb_generic_error_t *error = NULL;
+
+   xcb_shm_seg_t shm_seg = xcb_generate_id(dri2_dpy->conn);
+   error = xcb_request_check(dri2_dpy->conn,
+                             xcb_shm_attach_checked(dri2_dpy->conn,
+                                                    shm_seg, shmid, 0));
+   if (error) {
+      mesa_loge("Failed to attach to x11 shm");
+      _eglError(EGL_BAD_SURFACE, "xcb_shm_attach_checked");
+      free(error);
+      return;
+   }
+
+   xcb_gcontext_t gc;
+   xcb_void_cookie_t cookie;
+   switch (op) {
+   case __DRI_SWRAST_IMAGE_OP_DRAW:
+      gc = dri2_surf->gc;
+      break;
+   case __DRI_SWRAST_IMAGE_OP_SWAP:
+      gc = dri2_surf->swapgc;
+      break;
+   default:
+      return;
+   }
+
+   cookie = xcb_shm_put_image(dri2_dpy->conn,
+         dri2_surf->drawable,
+         gc,
+         stride / dri2_surf->bytes_per_pixel, h,
+         x, 0,
+         w, h,
+         x, y,
+         dri2_surf->depth,
+         XCB_IMAGE_FORMAT_Z_PIXMAP,
+         0, shm_seg, stride * y);
+   xcb_discard_reply(dri2_dpy->conn, cookie.sequence);
+
+   xcb_flush(dri2_dpy->conn);
+   xcb_shm_detach(dri2_dpy->conn, shm_seg);
+}
+
+static void
+swrastGetImageShm(__DRIdrawable * read,
+                  int x, int y, int w, int h,
+                  int shmid, void *loaderPrivate)
+{
+   struct dri2_egl_surface *dri2_surf = loaderPrivate;
+   struct dri2_egl_display *dri2_dpy =
+      dri2_egl_display(dri2_surf->base.Resource.Display);
+   xcb_generic_error_t *error = NULL;
+
+   xcb_shm_seg_t shm_seg = xcb_generate_id(dri2_dpy->conn);
+   error = xcb_request_check(dri2_dpy->conn,
+                             xcb_shm_attach_checked(dri2_dpy->conn,
+                                                    shm_seg, shmid, 0));
+   if (error) {
+      mesa_loge("Failed to attach to x11 shm");
+      _eglError(EGL_BAD_SURFACE, "xcb_shm_attach_checked");
+      free(error);
+      return;
+   }
+
+   xcb_shm_get_image_cookie_t cookie;
+   xcb_shm_get_image_reply_t *reply;
+
+   cookie = xcb_shm_get_image(dri2_dpy->conn,
+         dri2_surf->drawable,
+         x, y,
+         w, h,
+         ~0, XCB_IMAGE_FORMAT_Z_PIXMAP,
+         shm_seg, 0);
+   reply = xcb_shm_get_image_reply(dri2_dpy->conn, cookie, NULL);
+   if (reply == NULL)
+      _eglLog(_EGL_WARNING, "error in xcb_shm_get_image");
+   else
+      free(reply);
+
+   xcb_shm_detach(dri2_dpy->conn, shm_seg);
+}
+
 static xcb_screen_t *
 get_xcb_screen(xcb_screen_iterator_t iter, int screen)
 {
@@ -744,6 +883,11 @@ dri2_x11_connect(struct dri2_egl_display *dri2_dpy)
          strndup(driver_name, xcb_dri2_connect_driver_name_length(connect));
    }
 
+   if (!strcmp(dri2_dpy->driver_name, "zink")) {
+      close(dri2_dpy->fd_render_gpu);
+      return EGL_FALSE;
+   }
+
    if (dri2_dpy->driver_name == NULL) {
       close(dri2_dpy->fd_render_gpu);
       free(connect);
@@ -768,13 +912,12 @@ dri2_x11_authenticate(_EGLDisplay *disp, uint32_t id)
    return dri2_x11_do_authenticate(dri2_dpy, id);
 }
 
-static EGLBoolean
+static void
 dri2_x11_add_configs_for_visuals(struct dri2_egl_display *dri2_dpy,
                                  _EGLDisplay *disp, bool supports_preserved)
 {
    xcb_depth_iterator_t d;
    xcb_visualtype_t *visuals;
-   int config_count = 0;
    EGLint surface_type;
 
    d = xcb_screen_allowed_depths_iterator(dri2_dpy->screen);
@@ -797,72 +940,76 @@ dri2_x11_add_configs_for_visuals(struct dri2_egl_display *dri2_dpy,
 
          class_added[visuals[i]._class] = EGL_TRUE;
 
+         const int rgb_shifts[3] = {
+            ffs(visuals[i].red_mask) - 1,
+            ffs(visuals[i].green_mask) - 1,
+            ffs(visuals[i].blue_mask) - 1,
+         };
+
+         const unsigned int rgb_sizes[3] = {
+            util_bitcount(visuals[i].red_mask),
+            util_bitcount(visuals[i].green_mask),
+            util_bitcount(visuals[i].blue_mask),
+         };
+
+         const EGLint config_attrs[] = {
+            EGL_NATIVE_VISUAL_ID,
+            visuals[i].visual_id,
+            EGL_NATIVE_VISUAL_TYPE,
+            visuals[i]._class,
+            EGL_NONE,
+         };
+
+         const EGLint config_attrs_2nd_group[] = {
+            EGL_NATIVE_VISUAL_ID,
+            visuals[i].visual_id,
+            EGL_NATIVE_VISUAL_TYPE,
+            visuals[i]._class,
+            EGL_CONFIG_SELECT_GROUP_EXT,
+            1,
+            EGL_NONE,
+         };
+
          for (int j = 0; dri2_dpy->driver_configs[j]; j++) {
-            struct dri2_egl_config *dri2_conf;
             const __DRIconfig *config = dri2_dpy->driver_configs[j];
+            int shifts[4];
+            unsigned int sizes[4];
 
-            const EGLint config_attrs[] = {
-               EGL_NATIVE_VISUAL_ID,
-               visuals[i].visual_id,
-               EGL_NATIVE_VISUAL_TYPE,
-               visuals[i]._class,
-               EGL_NONE,
-            };
+            dri2_get_shifts_and_sizes(dri2_dpy->core, config, shifts, sizes);
 
-            int rgba_shifts[4] = {
-               ffs(visuals[i].red_mask) - 1,
-               ffs(visuals[i].green_mask) - 1,
-               ffs(visuals[i].blue_mask) - 1,
-               -1,
-            };
+            if (memcmp(shifts, rgb_shifts, sizeof(rgb_shifts)) != 0 ||
+                memcmp(sizes, rgb_sizes, sizeof(rgb_sizes)) != 0) {
+               continue;
+            }
 
-            unsigned int rgba_sizes[4] = {
-               util_bitcount(visuals[i].red_mask),
-               util_bitcount(visuals[i].green_mask),
-               util_bitcount(visuals[i].blue_mask),
-               0,
-            };
-
-            dri2_conf =
-               dri2_add_config(disp, config, config_count + 1, surface_type,
-                               config_attrs, rgba_shifts, rgba_sizes);
-            if (dri2_conf)
-               if (dri2_conf->base.ConfigID == config_count + 1)
-                  config_count++;
-
-            /* Allow a 24-bit RGB visual to match a 32-bit RGBA EGLConfig.
-             * Ditto for 30-bit RGB visuals to match a 32-bit RGBA EGLConfig.
+            /* Allows RGB visuals to match a 32-bit RGBA EGLConfig.
              * Otherwise it will only match a 32-bit RGBA visual.  On a
              * composited window manager on X11, this will make all of the
              * EGLConfigs with destination alpha get blended by the
              * compositor.  This is probably not what the application
              * wants... especially on drivers that only have 32-bit RGBA
              * EGLConfigs! */
-            if (d.data->depth == 24 || d.data->depth == 30) {
+            if (sizes[3] != 0) {
                unsigned int rgba_mask =
                   ~(visuals[i].red_mask | visuals[i].green_mask |
                     visuals[i].blue_mask);
-               rgba_shifts[3] = ffs(rgba_mask) - 1;
-               rgba_sizes[3] = util_bitcount(rgba_mask);
-               dri2_conf =
-                  dri2_add_config(disp, config, config_count + 1, surface_type,
-                                  config_attrs, rgba_shifts, rgba_sizes);
-               if (dri2_conf)
-                  if (dri2_conf->base.ConfigID == config_count + 1)
-                     config_count++;
+
+               if (shifts[3] != ffs(rgba_mask) - 1 ||
+                   sizes[3] != util_bitcount(rgba_mask))
+                  continue;
+            }
+
+            unsigned int bit_per_pixel = sizes[0] + sizes[1] + sizes[2] + sizes[3];
+            if (sizes[3] != 0 && d.data->depth == bit_per_pixel) {
+               dri2_add_config(disp, config, surface_type, config_attrs_2nd_group);
+            } else {
+               dri2_add_config(disp, config, surface_type, config_attrs);
             }
          }
       }
 
       xcb_depth_next(&d);
    }
-
-   if (!config_count) {
-      _eglLog(_EGL_WARNING, "DRI2: failed to create any config");
-      return EGL_FALSE;
-   }
-
-   return EGL_TRUE;
 }
 
 static EGLBoolean
@@ -1013,6 +1160,41 @@ dri2_x11_post_sub_buffer(_EGLDisplay *disp, _EGLSurface *draw, EGLint x,
    return dri2_x11_swap_buffers_region(disp, draw, 1, rect);
 }
 
+static EGLBoolean
+dri2_x11_kopper_swap_buffers_with_damage(_EGLDisplay *disp, _EGLSurface *draw,
+                                         const EGLint *rects, EGLint numRects)
+{
+   struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
+   struct dri2_egl_surface *dri2_surf = dri2_egl_surface(draw);
+   /* swrast path unsupported for now */
+   assert(dri2_dpy->kopper);
+   if (numRects) {
+      if (dri2_dpy->kopper)
+         dri2_dpy->kopper->swapBuffersWithDamage(dri2_surf->dri_drawable, __DRI2_FLUSH_INVALIDATE_ANCILLARY, numRects, rects);
+      else
+         dri2_dpy->core->swapBuffersWithDamage(dri2_surf->dri_drawable, numRects, rects);
+   } else {
+      if (dri2_dpy->kopper)
+         dri2_dpy->kopper->swapBuffers(dri2_surf->dri_drawable, __DRI2_FLUSH_INVALIDATE_ANCILLARY);
+      else
+         dri2_dpy->core->swapBuffers(dri2_surf->dri_drawable);
+   }
+   return EGL_TRUE;
+}
+
+static EGLBoolean
+dri2_x11_swap_buffers_with_damage(_EGLDisplay *disp, _EGLSurface *draw,
+                                  const EGLint *rects, EGLint numRects)
+{
+   struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
+   struct dri2_egl_surface *dri2_surf = dri2_egl_surface(draw);
+   if (numRects)
+      dri2_dpy->core->swapBuffersWithDamage(dri2_surf->dri_drawable, numRects, rects);
+   else
+      dri2_dpy->core->swapBuffers(dri2_surf->dri_drawable);
+   return EGL_TRUE;
+}
+
 static EGLBoolean
 dri2_x11_swap_interval(_EGLDisplay *disp, _EGLSurface *surf, EGLint interval)
 {
@@ -1064,23 +1246,23 @@ dri2_x11_copy_buffers(_EGLDisplay *disp, _EGLSurface *surf,
 }
 
 uint32_t
-dri2_format_for_depth(struct dri2_egl_display *dri2_dpy, uint32_t depth)
+dri2_fourcc_for_depth(struct dri2_egl_display *dri2_dpy, uint32_t depth)
 {
    switch (depth) {
    case 16:
-      return __DRI_IMAGE_FORMAT_RGB565;
+      return DRM_FORMAT_RGB565;
    case 24:
-      return __DRI_IMAGE_FORMAT_XRGB8888;
+      return DRM_FORMAT_XRGB8888;
    case 30:
       /* Different preferred formats for different hw */
       if (dri2_x11_get_red_mask_for_depth(dri2_dpy, 30) == 0x3ff)
-         return __DRI_IMAGE_FORMAT_XBGR2101010;
+         return DRM_FORMAT_XBGR2101010;
       else
-         return __DRI_IMAGE_FORMAT_XRGB2101010;
+         return DRM_FORMAT_XRGB2101010;
    case 32:
-      return __DRI_IMAGE_FORMAT_ARGB8888;
+      return DRM_FORMAT_ARGB8888;
    default:
-      return __DRI_IMAGE_FORMAT_NONE;
+      return DRM_FORMAT_INVALID;
    }
 }
 
@@ -1098,7 +1280,7 @@ dri2_create_image_khr_pixmap(_EGLDisplay *disp, _EGLContext *ctx,
    xcb_get_geometry_cookie_t geometry_cookie;
    xcb_get_geometry_reply_t *geometry_reply;
    xcb_generic_error_t *error;
-   int stride, format;
+   int fourcc;
 
    (void)ctx;
 
@@ -1129,8 +1311,8 @@ dri2_create_image_khr_pixmap(_EGLDisplay *disp, _EGLContext *ctx,
       return NULL;
    }
 
-   format = dri2_format_for_depth(dri2_dpy, geometry_reply->depth);
-   if (format == __DRI_IMAGE_FORMAT_NONE) {
+   fourcc = dri2_fourcc_for_depth(dri2_dpy, geometry_reply->depth);
+   if (fourcc == DRM_FORMAT_INVALID) {
       _eglError(EGL_BAD_PARAMETER,
                 "dri2_create_image_khr: unsupported pixmap depth");
       free(buffers_reply);
@@ -1148,10 +1330,10 @@ dri2_create_image_khr_pixmap(_EGLDisplay *disp, _EGLContext *ctx,
 
    _eglInitImage(&dri2_img->base, disp);
 
-   stride = buffers[0].pitch / buffers[0].cpp;
-   dri2_img->dri_image = dri2_dpy->image->createImageFromName(
+   dri2_img->dri_image = dri2_dpy->image->createImageFromNames(
       dri2_dpy->dri_screen_render_gpu, buffers_reply->width,
-      buffers_reply->height, format, buffers[0].name, stride, dri2_img);
+      buffers_reply->height, fourcc, (int *) &buffers[0].name, 1,
+      (int *) &buffers[0].pitch, 0, dri2_img);
 
    free(buffers_reply);
    free(geometry_reply);
@@ -1212,6 +1394,7 @@ dri2_x11_get_msc_rate(_EGLDisplay *display, _EGLSurface *surface,
 {
    struct dri2_egl_display *dri2_dpy = dri2_egl_display(display);
 
+#ifdef HAVE_DRI3
    loader_update_screen_resources(&dri2_dpy->screen_resources);
 
    if (dri2_dpy->screen_resources.num_crtcs == 0) {
@@ -1228,6 +1411,10 @@ dri2_x11_get_msc_rate(_EGLDisplay *display, _EGLSurface *surface,
    /* If there's only one active CRTC, we're done */
    if (dri2_dpy->screen_resources.num_crtcs == 1)
       return EGL_TRUE;
+#else
+   *numerator = 0;
+   *denominator = 1;
+#endif
 
    /* In a multi-monitor setup, look at each CRTC and perform a box
     * intersection between the CRTC and surface.  Use the CRTC whose
@@ -1250,6 +1437,7 @@ dri2_x11_get_msc_rate(_EGLDisplay *display, _EGLSurface *surface,
       return EGL_FALSE;
    }
 
+#ifdef HAVE_DRI3
    int area = 0;
 
    for (unsigned c = 0; c < dri2_dpy->screen_resources.num_crtcs; c++) {
@@ -1264,6 +1452,7 @@ dri2_x11_get_msc_rate(_EGLDisplay *display, _EGLSurface *surface,
          area = c_area;
       }
    }
+#endif
 
    /* If the window is entirely off-screen, then area will still be 0.
     * We defaulted to the first CRTC in the list's refresh rate, earlier.
@@ -1317,10 +1506,22 @@ dri2_kopper_query_buffer_age(_EGLDisplay *disp, _EGLSurface *surf)
    /* This can legitimately be null for lavapipe */
    if (dri2_dpy->kopper)
       return dri2_dpy->kopper->queryBufferAge(dri2_surf->dri_drawable);
+   else
+      return dri2_dpy->swrast->queryBufferAge(dri2_surf->dri_drawable);
 
    return 0;
 }
 
+static EGLint
+dri2_swrast_query_buffer_age(_EGLDisplay *disp, _EGLSurface *surf)
+{
+   struct dri2_egl_display *dri2_dpy = dri2_egl_display(disp);
+   struct dri2_egl_surface *dri2_surf = dri2_egl_surface(surf);
+
+   assert(dri2_dpy->swrast);
+   return dri2_dpy->swrast->queryBufferAge(dri2_surf->dri_drawable);
+}
+
 static const struct dri2_egl_display_vtbl dri2_x11_swrast_display_vtbl = {
    .authenticate = NULL,
    .create_window_surface = dri2_x11_create_window_surface,
@@ -1330,8 +1531,10 @@ static const struct dri2_egl_display_vtbl dri2_x11_swrast_display_vtbl = {
    .create_image = dri2_create_image_khr,
    .swap_buffers = dri2_x11_swap_buffers,
    .swap_buffers_region = dri2_x11_swap_buffers_region,
+   .swap_buffers_with_damage = dri2_x11_swap_buffers_with_damage,
    .post_sub_buffer = dri2_x11_post_sub_buffer,
    .copy_buffers = dri2_x11_copy_buffers,
+   .query_buffer_age = dri2_swrast_query_buffer_age,
    /* XXX: should really implement this since X11 has pixmaps */
    .query_surface = dri2_query_surface,
    .get_msc_rate = dri2_x11_get_msc_rate,
@@ -1348,6 +1551,7 @@ static const struct dri2_egl_display_vtbl dri2_x11_kopper_display_vtbl = {
    .swap_interval = dri2_kopper_swap_interval,
    .swap_buffers = dri2_x11_swap_buffers,
    .swap_buffers_region = dri2_x11_swap_buffers_region,
+   .swap_buffers_with_damage = dri2_x11_kopper_swap_buffers_with_damage,
    .post_sub_buffer = dri2_x11_post_sub_buffer,
    .copy_buffers = dri2_x11_copy_buffers,
    .query_buffer_age = dri2_kopper_query_buffer_age,
@@ -1380,9 +1584,22 @@ static const __DRIswrastLoaderExtension swrast_loader_extension = {
 
    .getDrawableInfo = swrastGetDrawableInfo,
    .putImage = swrastPutImage,
+   .putImage2 = swrastPutImage2,
    .getImage = swrastGetImage,
 };
 
+static const __DRIswrastLoaderExtension swrast_loader_shm_extension = {
+   .base = {__DRI_SWRAST_LOADER, 4},
+
+   .getDrawableInfo = swrastGetDrawableInfo,
+   .putImage = swrastPutImage,
+   .putImage2 = swrastPutImage2,
+   .putImageShm = swrastPutImageShm,
+   .getImage = swrastGetImage,
+   .getImage2 = swrastGetImage2,
+   .getImageShm = swrastGetImageShm,
+};
+
 static_assert(sizeof(struct kopper_vk_surface_create_storage) >=
                  sizeof(VkXcbSurfaceCreateInfoKHR),
               "");
@@ -1403,6 +1620,7 @@ kopperSetSurfaceCreateInfo(void *_draw, struct kopper_loader_info *ci)
    xcb->connection = dri2_dpy->conn;
    xcb->window = dri2_surf->drawable;
    ci->has_alpha = dri2_surf->depth == 32;
+   ci->present_opaque = dri2_surf->base.PresentOpaque;
 }
 
 static const __DRIkopperLoaderExtension kopper_loader_extension = {
@@ -1418,6 +1636,13 @@ static const __DRIextension *swrast_loader_extensions[] = {
    NULL,
 };
 
+static const __DRIextension *swrast_loader_shm_extensions[] = {
+   &swrast_loader_shm_extension.base,
+   &image_lookup_extension.base,
+   &kopper_loader_extension.base,
+   NULL,
+};
+
 static int
 dri2_find_screen_for_display(const _EGLDisplay *disp, int fallback_screen)
 {
@@ -1502,6 +1727,38 @@ dri2_x11_setup_swap_interval(_EGLDisplay *disp)
    dri2_setup_swap_interval(disp, arbitrary_max_interval);
 }
 
+static bool
+check_xshm(struct dri2_egl_display *dri2_dpy)
+{
+   xcb_void_cookie_t cookie;
+   xcb_generic_error_t *error;
+   int ret = true;
+   xcb_query_extension_cookie_t shm_cookie;
+   xcb_query_extension_reply_t *shm_reply;
+   bool has_mit_shm;
+
+   shm_cookie = xcb_query_extension(dri2_dpy->conn, 7, "MIT-SHM");
+   shm_reply = xcb_query_extension_reply(dri2_dpy->conn, shm_cookie, NULL);
+
+   has_mit_shm = shm_reply->present;
+   free(shm_reply);
+   if (!has_mit_shm)
+      return false;
+
+   cookie = xcb_shm_detach_checked(dri2_dpy->conn, 0);
+   if ((error = xcb_request_check(dri2_dpy->conn, cookie))) {
+      /* BadRequest means we're a remote client. If we were local we'd
+       * expect BadValue since 'info' has an invalid segment name.
+       */
+      if (error->error_code == BadRequest)
+         ret = false;
+      free(error);
+   }
+
+   return ret;
+}
+
+
 static EGLBoolean
 dri2_initialize_x11_swrast(_EGLDisplay *disp)
 {
@@ -1517,13 +1774,20 @@ dri2_initialize_x11_swrast(_EGLDisplay *disp)
     * here will allow is to simply free the memory at dri2_terminate().
     */
    dri2_dpy->driver_name = strdup(disp->Options.Zink ? "zink" : "swrast");
+#ifdef HAVE_DRI3
    if (disp->Options.Zink &&
-       !debug_get_bool_option("LIBGL_DRI3_DISABLE", false))
-      dri3_x11_connect(dri2_dpy);
+       !debug_get_bool_option("LIBGL_DRI3_DISABLE", false) &&
+       !debug_get_bool_option("LIBGL_KOPPER_DRI2", false))
+      dri3_x11_connect(dri2_dpy, disp->Options.Zink, disp->Options.ForceSoftware);
+#endif
    if (!dri2_load_driver_swrast(disp))
       goto cleanup;
 
-   dri2_dpy->loader_extensions = swrast_loader_extensions;
+   if (check_xshm(dri2_dpy)) {
+      dri2_dpy->loader_extensions = swrast_loader_shm_extensions;
+   } else {
+      dri2_dpy->loader_extensions = swrast_loader_extensions;
+   }
 
    if (!dri2_create_screen(disp))
       goto cleanup;
@@ -1549,19 +1813,19 @@ dri2_initialize_x11_swrast(_EGLDisplay *disp)
          disp->Extensions.KHR_image_pixmap = EGL_TRUE;
       disp->Extensions.NOK_texture_from_pixmap = EGL_TRUE;
       disp->Extensions.CHROMIUM_sync_control = EGL_TRUE;
-      disp->Extensions.ANGLE_sync_control_rate = EGL_TRUE;
-      disp->Extensions.EXT_buffer_age = EGL_TRUE;
-      disp->Extensions.EXT_swap_buffers_with_damage = EGL_TRUE;
+      disp->Extensions.EXT_swap_buffers_with_damage = !!dri2_dpy->kopper;
 
+#ifdef HAVE_DRI3
       if (dri2_dpy->multibuffers_available)
          dri2_set_WL_bind_wayland_display(disp);
+#endif
    } else {
-      /* swrast */
-      disp->Extensions.ANGLE_sync_control_rate = EGL_TRUE;
+      disp->Extensions.EXT_swap_buffers_with_damage = EGL_TRUE;
    }
+   disp->Extensions.EXT_buffer_age = EGL_TRUE;
+   disp->Extensions.ANGLE_sync_control_rate = EGL_TRUE;
 
-   if (!dri2_x11_add_configs_for_visuals(dri2_dpy, disp, !disp->Options.Zink))
-      goto cleanup;
+   dri2_x11_add_configs_for_visuals(dri2_dpy, disp, !disp->Options.Zink);
 
    /* Fill vtbl last to prevent accidentally calling virtual function during
     * initialization.
@@ -1588,18 +1852,19 @@ static const __DRIextension *dri3_image_loader_extensions[] = {
    NULL,
 };
 
-static EGLBoolean
+static enum dri2_egl_driver_fail
 dri2_initialize_x11_dri3(_EGLDisplay *disp)
 {
    struct dri2_egl_display *dri2_dpy = dri2_display_create();
-
+   enum dri2_egl_driver_fail status = DRI2_EGL_DRIVER_FAILED;
    if (!dri2_dpy)
-      return EGL_FALSE;
+      return DRI2_EGL_DRIVER_FAILED;
 
    if (!dri2_get_xcb_connection(disp, dri2_dpy))
       goto cleanup;
 
-   if (!dri3_x11_connect(dri2_dpy))
+   status = dri3_x11_connect(dri2_dpy, disp->Options.Zink, disp->Options.ForceSoftware);
+   if (status != DRI2_EGL_DRIVER_LOADED)
       goto cleanup;
 
    if (!dri2_load_driver_dri3(disp))
@@ -1635,8 +1900,7 @@ dri2_initialize_x11_dri3(_EGLDisplay *disp)
 
    dri2_set_WL_bind_wayland_display(disp);
 
-   if (!dri2_x11_add_configs_for_visuals(dri2_dpy, disp, false))
-      goto cleanup;
+   dri2_x11_add_configs_for_visuals(dri2_dpy, disp, false);
 
    loader_init_screen_resources(&dri2_dpy->screen_resources, dri2_dpy->conn,
                                 dri2_dpy->screen);
@@ -1655,11 +1919,13 @@ dri2_initialize_x11_dri3(_EGLDisplay *disp)
 
    _eglLog(_EGL_INFO, "Using DRI3");
 
-   return EGL_TRUE;
+   return DRI2_EGL_DRIVER_LOADED;
 
 cleanup:
    dri2_display_destroy(disp);
-   return EGL_FALSE;
+   return status == DRI2_EGL_DRIVER_PREFER_ZINK ?
+          DRI2_EGL_DRIVER_PREFER_ZINK :
+          DRI2_EGL_DRIVER_FAILED;
 }
 #endif
 
@@ -1742,8 +2008,7 @@ dri2_initialize_x11_dri2(_EGLDisplay *disp)
 
    dri2_set_WL_bind_wayland_display(disp);
 
-   if (!dri2_x11_add_configs_for_visuals(dri2_dpy, disp, true))
-      goto cleanup;
+   dri2_x11_add_configs_for_visuals(dri2_dpy, disp, true);
 
    /* Fill vtbl last to prevent accidentally calling virtual function during
     * initialization.
@@ -1762,16 +2027,21 @@ cleanup:
 EGLBoolean
 dri2_initialize_x11(_EGLDisplay *disp)
 {
-   if (disp->Options.ForceSoftware || disp->Options.Zink)
+   enum dri2_egl_driver_fail status = DRI2_EGL_DRIVER_FAILED;
+   if (disp->Options.ForceSoftware ||
+       (disp->Options.Zink && !debug_get_bool_option("LIBGL_KOPPER_DISABLE", false)))
       return dri2_initialize_x11_swrast(disp);
 
 #ifdef HAVE_DRI3
-   if (!debug_get_bool_option("LIBGL_DRI3_DISABLE", false))
-      if (dri2_initialize_x11_dri3(disp))
+   if (!debug_get_bool_option("LIBGL_DRI3_DISABLE", false)) {
+      status = dri2_initialize_x11_dri3(disp);
+      if (status == DRI2_EGL_DRIVER_LOADED)
          return EGL_TRUE;
+   }
 #endif
 
-   if (!debug_get_bool_option("LIBGL_DRI2_DISABLE", false))
+   if (!debug_get_bool_option("LIBGL_DRI2_DISABLE", false) &&
+       status != DRI2_EGL_DRIVER_PREFER_ZINK)
       if (dri2_initialize_x11_dri2(disp))
          return EGL_TRUE;
 
@@ -1781,8 +2051,10 @@ dri2_initialize_x11(_EGLDisplay *disp)
 void
 dri2_teardown_x11(struct dri2_egl_display *dri2_dpy)
 {
+#ifdef HAVE_DRI3
    if (dri2_dpy->dri2_major >= 3)
       loader_destroy_screen_resources(&dri2_dpy->screen_resources);
+#endif
 
    if (dri2_dpy->own_device)
       xcb_disconnect(dri2_dpy->conn);
