@@ -28,7 +28,9 @@
 #ifndef BRW_FS_H
 #define BRW_FS_H
 
-#include "brw_shader.h"
+#include "brw_cfg.h"
+#include "brw_compiler.h"
+#include "brw_ir_allocator.h"
 #include "brw_ir_fs.h"
 #include "brw_fs_live_variables.h"
 #include "brw_ir_performance.h"
@@ -39,7 +41,7 @@ namespace {
    struct acp_entry;
 }
 
-class fs_visitor;
+struct fs_visitor;
 
 namespace brw {
    /**
@@ -67,9 +69,74 @@ namespace brw {
 
       unsigned *regs_live_at_ip;
    };
+
+   class def_analysis {
+   public:
+      def_analysis(const fs_visitor *v);
+      ~def_analysis();
+
+      fs_inst *
+      get(const brw_reg &reg) const
+      {
+         return reg.file == VGRF && reg.nr < def_count ?
+                def_insts[reg.nr] : NULL;
+      }
+
+      bblock_t *
+      get_block(const brw_reg &reg) const
+      {
+         return reg.file == VGRF && reg.nr < def_count ?
+                def_blocks[reg.nr] : NULL;
+      }
+
+      uint32_t
+      get_use_count(const brw_reg &reg) const
+      {
+         return reg.file == VGRF && reg.nr < def_count ?
+                def_use_counts[reg.nr] : 0;
+      }
+
+      unsigned count() const { return def_count; }
+
+      void print_stats(const fs_visitor *) const;
+
+      analysis_dependency_class
+      dependency_class() const
+      {
+         return DEPENDENCY_INSTRUCTION_IDENTITY |
+                DEPENDENCY_INSTRUCTION_DATA_FLOW |
+                DEPENDENCY_VARIABLES |
+                DEPENDENCY_BLOCKS;
+      }
+
+      bool validate(const fs_visitor *) const;
+
+   private:
+      void mark_invalid(int);
+      bool fully_defines(const fs_visitor *v, fs_inst *);
+      void update_for_reads(const idom_tree &idom, bblock_t *block, fs_inst *);
+      void update_for_write(const fs_visitor *v, bblock_t *block, fs_inst *);
+
+      fs_inst **def_insts;
+      bblock_t **def_blocks;
+      uint32_t *def_use_counts;
+      unsigned def_count;
+   };
 }
 
-struct brw_gs_compile;
+#define UBO_START ((1 << 16) - 4)
+
+/**
+ * Scratch data used when compiling a GLSL geometry shader.
+ */
+struct brw_gs_compile
+{
+   struct brw_gs_prog_key key;
+   struct intel_vue_map input_vue_map;
+
+   unsigned control_data_bits_per_vertex;
+   unsigned control_data_header_size_bits;
+};
 
 namespace brw {
 class fs_builder;
@@ -97,38 +164,38 @@ protected:
 struct vs_thread_payload : public thread_payload {
    vs_thread_payload(const fs_visitor &v);
 
-   fs_reg urb_handles;
+   brw_reg urb_handles;
 };
 
 struct tcs_thread_payload : public thread_payload {
    tcs_thread_payload(const fs_visitor &v);
 
-   fs_reg patch_urb_output;
-   fs_reg primitive_id;
-   fs_reg icp_handle_start;
+   brw_reg patch_urb_output;
+   brw_reg primitive_id;
+   brw_reg icp_handle_start;
 };
 
 struct tes_thread_payload : public thread_payload {
    tes_thread_payload(const fs_visitor &v);
 
-   fs_reg patch_urb_input;
-   fs_reg primitive_id;
-   fs_reg coords[3];
-   fs_reg urb_output;
+   brw_reg patch_urb_input;
+   brw_reg primitive_id;
+   brw_reg coords[3];
+   brw_reg urb_output;
 };
 
 struct gs_thread_payload : public thread_payload {
    gs_thread_payload(fs_visitor &v);
 
-   fs_reg urb_handles;
-   fs_reg primitive_id;
-   fs_reg icp_handle_start;
+   brw_reg urb_handles;
+   brw_reg primitive_id;
+   brw_reg instance_id;
+   brw_reg icp_handle_start;
 };
 
 struct fs_thread_payload : public thread_payload {
    fs_thread_payload(const fs_visitor &v,
-                     bool &source_depth_to_render_target,
-                     bool &runtime_check_aads_emit);
+                     bool &source_depth_to_render_target);
 
    uint8_t subspan_coord_reg[2];
    uint8_t source_depth_reg[2];
@@ -137,49 +204,63 @@ struct fs_thread_payload : public thread_payload {
    uint8_t dest_depth_reg[2];
    uint8_t sample_pos_reg[2];
    uint8_t sample_mask_in_reg[2];
-   uint8_t depth_w_coef_reg;
    uint8_t barycentric_coord_reg[BRW_BARYCENTRIC_MODE_COUNT][2];
+
+   uint8_t depth_w_coef_reg;
+   uint8_t pc_bary_coef_reg;
+   uint8_t npc_bary_coef_reg;
+   uint8_t sample_offsets_reg;
 };
 
 struct cs_thread_payload : public thread_payload {
    cs_thread_payload(const fs_visitor &v);
 
-   void load_subgroup_id(const brw::fs_builder &bld, fs_reg &dest) const;
+   void load_subgroup_id(const brw::fs_builder &bld, brw_reg &dest) const;
+
+   brw_reg local_invocation_id[3];
 
 protected:
-   fs_reg subgroup_id_;
+   brw_reg subgroup_id_;
 };
 
 struct task_mesh_thread_payload : public cs_thread_payload {
    task_mesh_thread_payload(fs_visitor &v);
 
-   fs_reg extended_parameter_0;
-   fs_reg local_index;
-   fs_reg inline_parameter;
+   brw_reg extended_parameter_0;
+   brw_reg local_index;
+   brw_reg inline_parameter;
 
-   fs_reg urb_output;
+   brw_reg urb_output;
 
    /* URB to read Task memory inputs. Only valid for MESH stage. */
-   fs_reg task_urb_input;
+   brw_reg task_urb_input;
 };
 
 struct bs_thread_payload : public thread_payload {
    bs_thread_payload(const fs_visitor &v);
 
-   fs_reg global_arg_ptr;
-   fs_reg local_arg_ptr;
+   brw_reg global_arg_ptr;
+   brw_reg local_arg_ptr;
 
-   void load_shader_type(const brw::fs_builder &bld, fs_reg &dest) const;
+   void load_shader_type(const brw::fs_builder &bld, brw_reg &dest) const;
 };
 
-class fs_instruction_scheduler;
+enum instruction_scheduler_mode {
+   SCHEDULE_PRE,
+   SCHEDULE_PRE_NON_LIFO,
+   SCHEDULE_PRE_LIFO,
+   SCHEDULE_POST,
+   SCHEDULE_NONE,
+};
+
+class instruction_scheduler;
 
 /**
  * The fragment shader front-end.
  *
  * Translates either GLSL IR or Mesa IR (for ARB_fragment_program) into FS IR.
  */
-class fs_visitor : public backend_shader
+struct fs_visitor
 {
 public:
    fs_visitor(const struct brw_compiler *compiler,
@@ -209,17 +290,16 @@ public:
    void init();
    ~fs_visitor();
 
-   fs_reg vgrf(const glsl_type *const type);
    void import_uniforms(fs_visitor *v);
 
    void VARYING_PULL_CONSTANT_LOAD(const brw::fs_builder &bld,
-                                   const fs_reg &dst,
-                                   const fs_reg &surface,
-                                   const fs_reg &surface_handle,
-                                   const fs_reg &varying_offset,
+                                   const brw_reg &dst,
+                                   const brw_reg &surface,
+                                   const brw_reg &surface_handle,
+                                   const brw_reg &varying_offset,
                                    uint32_t const_offset,
-                                   uint8_t alignment);
-   void DEP_RESOLVE_MOV(const brw::fs_builder &bld, int grf);
+                                   uint8_t alignment,
+                                   unsigned components);
 
    bool run_fs(bool allow_spilling, bool do_rep_send);
    bool run_vs();
@@ -230,14 +310,8 @@ public:
    bool run_bs(bool allow_spilling);
    bool run_task(bool allow_spilling);
    bool run_mesh(bool allow_spilling);
-   void optimize();
    void allocate_registers(bool allow_spilling);
    uint32_t compute_max_register_pressure();
-   bool fixup_sends_duplicate_payload();
-   void fixup_3src_null_dest();
-   void emit_dummy_memory_fence_before_eot();
-   void emit_dummy_mov_instruction();
-   bool fixup_nomask_control_flow();
    void assign_curb_setup();
    void assign_urb_setup();
    void convert_attr_sources_to_hw_regs(fs_inst *inst);
@@ -249,105 +323,86 @@ public:
    void assign_regs_trivial();
    void calculate_payload_ranges(unsigned payload_node_count,
                                  int *payload_last_use_ip) const;
-   bool split_virtual_grfs();
-   bool compact_virtual_grfs();
    void assign_constant_locations();
-   bool get_pull_locs(const fs_reg &src, unsigned *out_surf_index,
+   bool get_pull_locs(const brw_reg &src, unsigned *out_surf_index,
                       unsigned *out_pull_index);
-   bool lower_constant_loads();
-   virtual void invalidate_analysis(brw::analysis_dependency_class c);
+   void invalidate_analysis(brw::analysis_dependency_class c);
 
-#ifndef NDEBUG
-   void validate();
-#else
-   void validate() {}
-#endif
-
-   bool opt_algebraic();
-   bool opt_redundant_halt();
-   bool opt_cse();
-   bool opt_cse_local(const brw::fs_live_variables &live, bblock_t *block, int &ip);
-
-   bool opt_copy_propagation();
-   bool opt_bank_conflicts();
-   bool opt_split_sends();
-   bool register_coalesce();
-   bool compute_to_mrf();
-   bool eliminate_find_live_channel();
-   bool dead_code_eliminate();
-   bool remove_duplicate_mrf_writes();
-   bool remove_extra_rounding_modes();
-
-   fs_instruction_scheduler *prepare_scheduler(void *mem_ctx);
-   void schedule_instructions_pre_ra(fs_instruction_scheduler *sched,
+   instruction_scheduler *prepare_scheduler(void *mem_ctx);
+   void schedule_instructions_pre_ra(instruction_scheduler *sched,
                                      instruction_scheduler_mode mode);
    void schedule_instructions_post_ra();
 
-   void insert_gfx4_send_dependency_workarounds();
-   void insert_gfx4_pre_send_dependency_workarounds(bblock_t *block,
-                                                    fs_inst *inst);
-   void insert_gfx4_post_send_dependency_workarounds(bblock_t *block,
-                                                     fs_inst *inst);
    void vfail(const char *msg, va_list args);
    void fail(const char *msg, ...);
    void limit_dispatch_width(unsigned n, const char *msg);
-   bool lower_uniform_pull_constant_loads();
-   bool lower_load_payload();
-   bool lower_pack();
-   bool lower_regioning();
-   bool lower_logical_sends();
-   bool lower_integer_multiplication();
-   bool lower_minmax();
-   bool lower_simd_width();
-   bool lower_barycentrics();
-   bool lower_derivatives();
-   bool lower_find_live_channel();
-   bool lower_scoreboard();
-   bool lower_sub_sat();
-   bool opt_combine_constants();
 
    void emit_repclear_shader();
-   void emit_interpolation_setup_gfx4();
-   void emit_interpolation_setup_gfx6();
-   bool opt_peephole_sel();
-   bool opt_saturate_propagation();
-   bool opt_cmod_propagation();
-   bool opt_zero_samples();
+   void emit_interpolation_setup();
 
    void set_tcs_invocation_id();
 
-   void emit_alpha_test();
    fs_inst *emit_single_fb_write(const brw::fs_builder &bld,
-                                 fs_reg color1, fs_reg color2,
-                                 fs_reg src0_alpha, unsigned components);
+                                 brw_reg color1, brw_reg color2,
+                                 brw_reg src0_alpha, unsigned components);
    void do_emit_fb_writes(int nr_color_regions, bool replicate_alpha);
    void emit_fb_writes();
-   void emit_urb_writes(const fs_reg &gs_vertex_count = fs_reg());
-   void emit_gs_control_data_bits(const fs_reg &vertex_count);
+   void emit_urb_writes(const brw_reg &gs_vertex_count = brw_reg());
+   void emit_gs_control_data_bits(const brw_reg &vertex_count);
+   brw_reg gs_urb_channel_mask(const brw_reg &dword_index);
+   brw_reg gs_urb_per_slot_dword_index(const brw_reg &vertex_count);
    void emit_gs_thread_end();
    bool mark_last_urb_write_with_eot();
    void emit_tcs_thread_end();
    void emit_urb_fence();
    void emit_cs_terminate();
 
-   fs_reg interp_reg(const brw::fs_builder &bld, unsigned location,
+   brw_reg interp_reg(const brw::fs_builder &bld, unsigned location,
                      unsigned channel, unsigned comp);
-   fs_reg per_primitive_reg(const brw::fs_builder &bld,
+   brw_reg per_primitive_reg(const brw::fs_builder &bld,
                             int location, unsigned comp);
 
-   virtual void dump_instruction_to_file(const backend_instruction *inst, FILE *file) const;
-   virtual void dump_instructions_to_file(FILE *file) const;
+   void dump_instruction_to_file(const fs_inst *inst, FILE *file, const brw::def_analysis *defs) const;
+   void dump_instructions_to_file(FILE *file) const;
+
+   /* Convenience functions based on the above. */
+   void dump_instruction(const fs_inst *inst, FILE *file = stderr, const brw::def_analysis *defs = nullptr) const {
+      dump_instruction_to_file(inst, file, defs);
+   }
+   void dump_instructions(const char *name = nullptr) const;
+
+   void calculate_cfg();
+
+   const struct brw_compiler *compiler;
+   void *log_data; /* Passed to compiler->*_log functions */
+
+   const struct intel_device_info * const devinfo;
+   const nir_shader *nir;
+
+   /** ralloc context for temporary data used during compile */
+   void *mem_ctx;
+
+   /** List of fs_inst. */
+   exec_list instructions;
+
+   cfg_t *cfg;
+
+   gl_shader_stage stage;
+   bool debug_enabled;
+
+   brw::simple_allocator alloc;
 
    const brw_base_prog_key *const key;
-   const struct brw_sampler_prog_key_data *key_tex;
 
    struct brw_gs_compile *gs_compile;
 
    struct brw_stage_prog_data *prog_data;
 
-   brw_analysis<brw::fs_live_variables, backend_shader> live_analysis;
+   brw_analysis<brw::fs_live_variables, fs_visitor> live_analysis;
    brw_analysis<brw::register_pressure, fs_visitor> regpressure_analysis;
    brw_analysis<brw::performance, fs_visitor> performance_analysis;
+   brw_analysis<brw::idom_tree, fs_visitor> idom_analysis;
+   brw_analysis<brw::def_analysis, fs_visitor> def_analysis;
 
    /** Number of uniform variable components visited. */
    unsigned uniforms;
@@ -361,14 +416,12 @@ public:
     */
    int *push_constant_loc;
 
-   fs_reg frag_depth;
-   fs_reg frag_stencil;
-   fs_reg sample_mask;
-   fs_reg outputs[VARYING_SLOT_MAX];
-   fs_reg dual_src_output;
+   brw_reg frag_depth;
+   brw_reg frag_stencil;
+   brw_reg sample_mask;
+   brw_reg outputs[VARYING_SLOT_MAX];
+   brw_reg dual_src_output;
    int first_non_payload_grf;
-   /** Either BRW_MAX_GRF or GFX7_MRF_HACK_START */
-   unsigned max_grf;
 
    bool failed;
    char *fail_msg;
@@ -404,6 +457,11 @@ public:
       return *static_cast<fs_thread_payload *>(this->payload_);
    };
 
+   const fs_thread_payload &fs_payload() const {
+      assert(stage == MESA_SHADER_FRAGMENT);
+      return *static_cast<const fs_thread_payload *>(this->payload_);
+   };
+
    cs_thread_payload &cs_payload() {
       assert(gl_shader_stage_uses_workgroup(stage));
       return *static_cast<cs_thread_payload *>(this->payload_);
@@ -420,17 +478,16 @@ public:
    }
 
    bool source_depth_to_render_target;
-   bool runtime_check_aads_emit;
 
-   fs_reg pixel_x;
-   fs_reg pixel_y;
-   fs_reg pixel_z;
-   fs_reg wpos_w;
-   fs_reg pixel_w;
-   fs_reg delta_xy[BRW_BARYCENTRIC_MODE_COUNT];
-   fs_reg final_gs_vertex_count;
-   fs_reg control_data_bits;
-   fs_reg invocation_id;
+   brw_reg pixel_x;
+   brw_reg pixel_y;
+   brw_reg pixel_z;
+   brw_reg wpos_w;
+   brw_reg pixel_w;
+   brw_reg delta_xy[BRW_BARYCENTRIC_MODE_COUNT];
+   brw_reg final_gs_vertex_count;
+   brw_reg control_data_bits;
+   brw_reg invocation_id;
 
    unsigned grf_used;
    bool spilled_any_registers;
@@ -445,10 +502,6 @@ public:
 
    struct shader_stats shader_stats;
 
-   void lower_mul_dword_inst(fs_inst *inst, bblock_t *block);
-   void lower_mul_qword_inst(fs_inst *inst, bblock_t *block);
-   void lower_mulh_inst(fs_inst *inst, bblock_t *block);
-
    unsigned workgroup_size() const;
 
    void debug_optimizer(const nir_shader *nir,
@@ -456,17 +509,18 @@ public:
                         int iteration, int pass_num) const;
 };
 
+void brw_print_swsb(FILE *f, const struct intel_device_info *devinfo, const tgl_swsb swsb);
+
 /**
  * Return the flag register used in fragment shaders to keep track of live
  * samples.  On Gfx7+ we use f1.0-f1.1 to allow discard jumps in SIMD32
- * dispatch mode, while earlier generations are constrained to f0.1, which
- * limits the dispatch width to SIMD16 for fragment shaders that use discard.
+ * dispatch mode.
  */
 static inline unsigned
 sample_mask_flag_subreg(const fs_visitor &s)
 {
    assert(s.stage == MESA_SHADER_FRAGMENT);
-   return s.devinfo->ver >= 7 ? 2 : 1;
+   return 2;
 }
 
 /**
@@ -480,7 +534,6 @@ public:
    fs_generator(const struct brw_compiler *compiler,
                 const struct brw_compile_params *params,
                 struct brw_stage_prog_data *prog_data,
-                bool runtime_check_aads_emit,
                 gl_shader_stage stage);
    ~fs_generator();
 
@@ -495,45 +548,18 @@ public:
    const unsigned *get_assembly();
 
 private:
-   void fire_fb_write(fs_inst *inst,
-                      struct brw_reg payload,
-                      struct brw_reg implied_header,
-                      GLuint nr);
    void generate_send(fs_inst *inst,
                       struct brw_reg dst,
                       struct brw_reg desc,
                       struct brw_reg ex_desc,
                       struct brw_reg payload,
                       struct brw_reg payload2);
-   void generate_fb_write(fs_inst *inst, struct brw_reg payload);
-   void generate_fb_read(fs_inst *inst, struct brw_reg dst,
-                         struct brw_reg payload);
-   void generate_cs_terminate(fs_inst *inst, struct brw_reg payload);
    void generate_barrier(fs_inst *inst, struct brw_reg src);
-   bool generate_linterp(fs_inst *inst, struct brw_reg dst,
-			 struct brw_reg *src);
-   void generate_tex(fs_inst *inst, struct brw_reg dst,
-                     struct brw_reg surface_index,
-                     struct brw_reg sampler_index);
    void generate_ddx(const fs_inst *inst,
                      struct brw_reg dst, struct brw_reg src);
    void generate_ddy(const fs_inst *inst,
                      struct brw_reg dst, struct brw_reg src);
-   void generate_scratch_write(fs_inst *inst, struct brw_reg src);
-   void generate_scratch_read(fs_inst *inst, struct brw_reg dst);
-   void generate_scratch_read_gfx7(fs_inst *inst, struct brw_reg dst);
    void generate_scratch_header(fs_inst *inst, struct brw_reg dst);
-   void generate_uniform_pull_constant_load(fs_inst *inst, struct brw_reg dst,
-                                            struct brw_reg index,
-                                            struct brw_reg offset);
-   void generate_varying_pull_constant_load_gfx4(fs_inst *inst,
-                                                 struct brw_reg dst,
-                                                 struct brw_reg index);
-
-   void generate_set_sample_id(fs_inst *inst,
-                               struct brw_reg dst,
-                               struct brw_reg src0,
-                               struct brw_reg src1);
 
    void generate_halt(fs_inst *inst);
 
@@ -564,7 +590,6 @@ private:
    unsigned dispatch_width; /**< 8, 16 or 32 */
 
    exec_list discard_halt_patches;
-   bool runtime_check_aads_emit;
    bool debug_flag;
    const char *shader_name;
    gl_shader_stage stage;
@@ -572,46 +597,37 @@ private:
 };
 
 namespace brw {
-   fs_reg
+   brw_reg
    fetch_payload_reg(const brw::fs_builder &bld, uint8_t regs[2],
-                     brw_reg_type type = BRW_REGISTER_TYPE_F,
+                     brw_reg_type type = BRW_TYPE_F,
                      unsigned n = 1);
 
-   fs_reg
+   brw_reg
    fetch_barycentric_reg(const brw::fs_builder &bld, uint8_t regs[2]);
 
-   inline fs_reg
+   inline brw_reg
    dynamic_msaa_flags(const struct brw_wm_prog_data *wm_prog_data)
    {
-      return fs_reg(UNIFORM, wm_prog_data->msaa_flags_param,
-                    BRW_REGISTER_TYPE_UD);
+      return brw_uniform_reg(wm_prog_data->msaa_flags_param, BRW_TYPE_UD);
    }
 
    void
    check_dynamic_msaa_flag(const fs_builder &bld,
                            const struct brw_wm_prog_data *wm_prog_data,
-                           enum brw_wm_msaa_flags flag);
+                           enum intel_msaa_flags flag);
 
    bool
    lower_src_modifiers(fs_visitor *v, bblock_t *block, fs_inst *inst, unsigned i);
 }
 
 void shuffle_from_32bit_read(const brw::fs_builder &bld,
-                             const fs_reg &dst,
-                             const fs_reg &src,
+                             const brw_reg &dst,
+                             const brw_reg &src,
                              uint32_t first_component,
                              uint32_t components);
 
-fs_reg setup_imm_df(const brw::fs_builder &bld,
-                    double v);
-
-fs_reg setup_imm_b(const brw::fs_builder &bld,
-                   int8_t v);
-
-fs_reg setup_imm_ub(const brw::fs_builder &bld,
-                   uint8_t v);
-
-enum brw_barycentric_mode brw_barycentric_mode(nir_intrinsic_instr *intr);
+enum brw_barycentric_mode brw_barycentric_mode(const struct brw_wm_prog_key *key,
+                                               nir_intrinsic_instr *intr);
 
 uint32_t brw_fb_write_msg_control(const fs_inst *inst,
                                   const struct brw_wm_prog_data *prog_data);
@@ -620,14 +636,71 @@ void brw_compute_urb_setup_index(struct brw_wm_prog_data *wm_prog_data);
 
 bool brw_nir_lower_simd(nir_shader *nir, unsigned dispatch_width);
 
-fs_reg brw_sample_mask_reg(const brw::fs_builder &bld);
+brw_reg brw_sample_mask_reg(const brw::fs_builder &bld);
 void brw_emit_predicate_on_sample_mask(const brw::fs_builder &bld, fs_inst *inst);
 
 int brw_get_subgroup_id_param_index(const intel_device_info *devinfo,
                                     const brw_stage_prog_data *prog_data);
 
-bool brw_lower_dpas(fs_visitor &v);
-
 void nir_to_brw(fs_visitor *s);
 
+#ifndef NDEBUG
+void brw_fs_validate(const fs_visitor &s);
+#else
+static inline void brw_fs_validate(const fs_visitor &s) {}
+#endif
+
+void brw_fs_optimize(fs_visitor &s);
+
+bool brw_fs_lower_3src_null_dest(fs_visitor &s);
+bool brw_fs_lower_alu_restrictions(fs_visitor &s);
+bool brw_fs_lower_barycentrics(fs_visitor &s);
+bool brw_fs_lower_constant_loads(fs_visitor &s);
+bool brw_fs_lower_derivatives(fs_visitor &s);
+bool brw_fs_lower_dpas(fs_visitor &s);
+bool brw_fs_lower_find_live_channel(fs_visitor &s);
+bool brw_fs_lower_integer_multiplication(fs_visitor &s);
+bool brw_fs_lower_load_subgroup_invocation(fs_visitor &s);
+bool brw_fs_lower_indirect_mov(fs_visitor &s);
+bool brw_fs_lower_logical_sends(fs_visitor &s);
+bool brw_fs_lower_pack(fs_visitor &s);
+bool brw_fs_lower_load_payload(fs_visitor &s);
+bool brw_fs_lower_regioning(fs_visitor &s);
+bool brw_fs_lower_scoreboard(fs_visitor &s);
+bool brw_fs_lower_sends_overlapping_payload(fs_visitor &s);
+bool brw_fs_lower_simd_width(fs_visitor &s);
+bool brw_fs_lower_csel(fs_visitor &s);
+bool brw_fs_lower_sub_sat(fs_visitor &s);
+bool brw_fs_lower_uniform_pull_constant_loads(fs_visitor &s);
+void brw_fs_lower_vgrfs_to_fixed_grfs(fs_visitor &s);
+
+bool brw_fs_opt_algebraic(fs_visitor &s);
+bool brw_fs_opt_bank_conflicts(fs_visitor &s);
+bool brw_fs_opt_cmod_propagation(fs_visitor &s);
+bool brw_fs_opt_combine_constants(fs_visitor &s);
+bool brw_fs_opt_compact_virtual_grfs(fs_visitor &s);
+bool brw_fs_opt_copy_propagation(fs_visitor &s);
+bool brw_fs_opt_copy_propagation_defs(fs_visitor &s);
+bool brw_fs_opt_cse_defs(fs_visitor &s);
+bool brw_fs_opt_dead_code_eliminate(fs_visitor &s);
+bool brw_fs_opt_dead_control_flow_eliminate(fs_visitor &s);
+bool brw_fs_opt_eliminate_find_live_channel(fs_visitor &s);
+bool brw_fs_opt_peephole_sel(fs_visitor &s);
+bool brw_fs_opt_predicated_break(fs_visitor &s);
+bool brw_fs_opt_register_coalesce(fs_visitor &s);
+bool brw_fs_opt_remove_extra_rounding_modes(fs_visitor &s);
+bool brw_fs_opt_remove_redundant_halts(fs_visitor &s);
+bool brw_fs_opt_saturate_propagation(fs_visitor &s);
+bool brw_fs_opt_split_sends(fs_visitor &s);
+bool brw_fs_opt_split_virtual_grfs(fs_visitor &s);
+bool brw_fs_opt_zero_samples(fs_visitor &s);
+
+bool brw_fs_workaround_emit_dummy_mov_instruction(fs_visitor &s);
+bool brw_fs_workaround_memory_fence_before_eot(fs_visitor &s);
+bool brw_fs_workaround_nomask_control_flow(fs_visitor &s);
+
+/* Helpers. */
+unsigned brw_fs_get_lowered_simd_width(const fs_visitor *shader,
+                                       const fs_inst *inst);
+
 #endif /* BRW_FS_H */
