@@ -112,6 +112,48 @@ nvk_get_nir_options(struct vk_physical_device *vk_pdev,
       return nvk_cg_nir_options(pdev, stage);
 }
 
+nir_address_format
+nvk_ubo_addr_format(const struct nvk_physical_device *pdev,
+                    const struct vk_pipeline_robustness_state *rs)
+{
+   if (nvk_use_bindless_cbuf(&pdev->info)) {
+      return nir_address_format_vec2_index_32bit_offset;
+   } else if (rs->null_uniform_buffer_descriptor) {
+      /* We need bounds checking for null descriptors */
+      return nir_address_format_64bit_bounded_global;
+   } else {
+      switch (rs->uniform_buffers) {
+      case VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_DISABLED_EXT:
+         return nir_address_format_64bit_global_32bit_offset;
+      case VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_EXT:
+      case VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2_EXT:
+         return nir_address_format_64bit_bounded_global;
+      default:
+         unreachable("Invalid robust buffer access behavior");
+      }
+   }
+}
+
+nir_address_format
+nvk_ssbo_addr_format(const struct nvk_physical_device *pdev,
+                    const struct vk_pipeline_robustness_state *rs)
+{
+   if (rs->null_storage_buffer_descriptor) {
+      /* We need bounds checking for null descriptors */
+      return nir_address_format_64bit_bounded_global;
+   } else {
+      switch (rs->storage_buffers) {
+      case VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_DISABLED_EXT:
+         return nir_address_format_64bit_global_32bit_offset;
+      case VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_EXT:
+      case VK_PIPELINE_ROBUSTNESS_BUFFER_BEHAVIOR_ROBUST_BUFFER_ACCESS_2_EXT:
+         return nir_address_format_64bit_bounded_global;
+      default:
+         unreachable("Invalid robust buffer access behavior");
+      }
+   }
+}
+
 static struct spirv_to_nir_options
 nvk_get_spirv_options(struct vk_physical_device *vk_pdev,
                       UNUSED gl_shader_stage stage,
@@ -121,55 +163,9 @@ nvk_get_spirv_options(struct vk_physical_device *vk_pdev,
       container_of(vk_pdev, struct nvk_physical_device, vk);
 
    return (struct spirv_to_nir_options) {
-      .caps = {
-         .demote_to_helper_invocation = true,
-         .descriptor_array_dynamic_indexing = true,
-         .descriptor_array_non_uniform_indexing = true,
-         .descriptor_indexing = true,
-         .device_group = true,
-         .draw_parameters = true,
-         .float_controls = true,
-         .float16 = true,
-         .float64 = true,
-         .fragment_barycentric = true,
-         .geometry_streams = true,
-         .image_atomic_int64 = true,
-         .image_ms_array = true,
-         .image_read_without_format = true,
-         .image_write_without_format = true,
-         .int8 = true,
-         .int16 = true,
-         .int64 = true,
-         .int64_atomics = true,
-         .min_lod = true,
-         .multiview = true,
-         .physical_storage_buffer_address = true,
-         .runtime_descriptor_array = true,
-         .shader_clock = true,
-         .shader_sm_builtins_nv = true,
-         .shader_viewport_index_layer = true,
-         .sparse_residency = true,
-         .storage_8bit = true,
-         .storage_16bit = true,
-         .storage_image_ms = true,
-         .subgroup_arithmetic = true,
-         .subgroup_ballot = true,
-         .subgroup_basic = true,
-         .subgroup_quad = true,
-         .subgroup_rotate = true,
-         .subgroup_shuffle = true,
-         .subgroup_uniform_control_flow = true,
-         .subgroup_vote = true,
-         .tessellation = true,
-         .transform_feedback = true,
-         .variable_pointers = true,
-         .vk_memory_model_device_scope = true,
-         .vk_memory_model = true,
-         .workgroup_memory_explicit_layout = true,
-      },
-      .ssbo_addr_format = nvk_buffer_addr_format(rs->storage_buffers),
+      .ssbo_addr_format = nvk_ssbo_addr_format(pdev, rs),
       .phys_ssbo_addr_format = nir_address_format_64bit_global,
-      .ubo_addr_format = nvk_buffer_addr_format(rs->uniform_buffers),
+      .ubo_addr_format = nvk_ubo_addr_format(pdev, rs),
       .shared_addr_format = nir_address_format_32bit_offset,
       .min_ssbo_alignment = NVK_MIN_SSBO_ALIGNMENT,
       .min_ubo_alignment = nvk_min_cbuf_alignment(&pdev->info),
@@ -200,6 +196,11 @@ nvk_populate_fs_key(struct nak_fs_key *key,
    key->sample_locations_cb = 0;
    key->sample_locations_offset = nvk_root_descriptor_offset(draw.sample_locations);
 
+   /* Turn underestimate on when no state is availaible or if explicitly set */
+   if (state == NULL || state->rs == NULL ||
+       state->rs->conservative_mode == VK_CONSERVATIVE_RASTERIZATION_MODE_UNDERESTIMATE_EXT)
+      key->uses_underestimate = true;
+
    if (state == NULL)
       return;
 
@@ -256,54 +257,84 @@ nvk_hash_graphics_state(struct vk_physical_device *device,
 }
 
 static bool
-lower_load_global_constant_offset_instr(nir_builder *b,
-                                        nir_intrinsic_instr *intrin,
-                                        UNUSED void *_data)
+lower_load_intrinsic(nir_builder *b, nir_intrinsic_instr *load,
+                     UNUSED void *_data)
 {
-   if (intrin->intrinsic != nir_intrinsic_load_global_constant_offset &&
-       intrin->intrinsic != nir_intrinsic_load_global_constant_bounded)
+   switch (load->intrinsic) {
+   case nir_intrinsic_load_ubo: {
+      b->cursor = nir_before_instr(&load->instr);
+
+      nir_def *index = load->src[0].ssa;
+      nir_def *offset = load->src[1].ssa;
+      const enum gl_access_qualifier access = nir_intrinsic_access(load);
+      const uint32_t align_mul = nir_intrinsic_align_mul(load);
+      const uint32_t align_offset = nir_intrinsic_align_offset(load);
+
+      nir_def *val;
+      if (load->src[0].ssa->num_components == 1) {
+         val = nir_ldc_nv(b, load->num_components, load->def.bit_size,
+                           index, offset, .access = access,
+                           .align_mul = align_mul,
+                           .align_offset = align_offset);
+      } else if (load->src[0].ssa->num_components == 2) {
+         nir_def *handle = nir_pack_64_2x32(b, load->src[0].ssa);
+         val = nir_ldcx_nv(b, load->num_components, load->def.bit_size,
+                           handle, offset, .access = access,
+                           .align_mul = align_mul,
+                           .align_offset = align_offset);
+      } else {
+         unreachable("Invalid UBO index");
+      }
+      nir_def_rewrite_uses(&load->def, val);
+      return true;
+   }
+
+   case nir_intrinsic_load_global_constant_offset:
+   case nir_intrinsic_load_global_constant_bounded: {
+      b->cursor = nir_before_instr(&load->instr);
+
+      nir_def *base_addr = load->src[0].ssa;
+      nir_def *offset = load->src[1].ssa;
+
+      nir_def *zero = NULL;
+      if (load->intrinsic == nir_intrinsic_load_global_constant_bounded) {
+         nir_def *bound = load->src[2].ssa;
+
+         unsigned bit_size = load->def.bit_size;
+         assert(bit_size >= 8 && bit_size % 8 == 0);
+         unsigned byte_size = bit_size / 8;
+
+         zero = nir_imm_zero(b, load->num_components, bit_size);
+
+         unsigned load_size = byte_size * load->num_components;
+
+         nir_def *sat_offset =
+            nir_umin(b, offset, nir_imm_int(b, UINT32_MAX - (load_size - 1)));
+         nir_def *in_bounds =
+            nir_ilt(b, nir_iadd_imm(b, sat_offset, load_size - 1), bound);
+
+         nir_push_if(b, in_bounds);
+      }
+
+      nir_def *val =
+         nir_build_load_global_constant(b, load->def.num_components,
+                                        load->def.bit_size,
+                                        nir_iadd(b, base_addr, nir_u2u64(b, offset)),
+                                        .align_mul = nir_intrinsic_align_mul(load),
+                                        .align_offset = nir_intrinsic_align_offset(load));
+
+      if (load->intrinsic == nir_intrinsic_load_global_constant_bounded) {
+         nir_pop_if(b, NULL);
+         val = nir_if_phi(b, val, zero);
+      }
+
+      nir_def_rewrite_uses(&load->def, val);
+      return true;
+   }
+
+   default:
       return false;
-
-   b->cursor = nir_before_instr(&intrin->instr);
-
-   nir_def *base_addr = intrin->src[0].ssa;
-   nir_def *offset = intrin->src[1].ssa;
-
-   nir_def *zero = NULL;
-   if (intrin->intrinsic == nir_intrinsic_load_global_constant_bounded) {
-      nir_def *bound = intrin->src[2].ssa;
-
-      unsigned bit_size = intrin->def.bit_size;
-      assert(bit_size >= 8 && bit_size % 8 == 0);
-      unsigned byte_size = bit_size / 8;
-
-      zero = nir_imm_zero(b, intrin->num_components, bit_size);
-
-      unsigned load_size = byte_size * intrin->num_components;
-
-      nir_def *sat_offset =
-         nir_umin(b, offset, nir_imm_int(b, UINT32_MAX - (load_size - 1)));
-      nir_def *in_bounds =
-         nir_ilt(b, nir_iadd_imm(b, sat_offset, load_size - 1), bound);
-
-      nir_push_if(b, in_bounds);
    }
-
-   nir_def *val =
-      nir_build_load_global_constant(b, intrin->def.num_components,
-                                     intrin->def.bit_size,
-                                     nir_iadd(b, base_addr, nir_u2u64(b, offset)),
-                                     .align_mul = nir_intrinsic_align_mul(intrin),
-                                     .align_offset = nir_intrinsic_align_offset(intrin));
-
-   if (intrin->intrinsic == nir_intrinsic_load_global_constant_bounded) {
-      nir_pop_if(b, NULL);
-      val = nir_if_phi(b, val, zero);
-   }
-
-   nir_def_rewrite_uses(&intrin->def, val);
-
-   return true;
 }
 
 struct lower_ycbcr_state {
@@ -434,16 +465,16 @@ nvk_lower_nir(struct nvk_device *dev, nir_shader *nir,
       };
    }
 
-   NIR_PASS(_, nir, nvk_nir_lower_descriptors, rs,
+   NIR_PASS(_, nir, nvk_nir_lower_descriptors, pdev, rs,
             set_layout_count, set_layouts, cbuf_map);
    NIR_PASS(_, nir, nir_lower_explicit_io, nir_var_mem_global,
             nir_address_format_64bit_global);
    NIR_PASS(_, nir, nir_lower_explicit_io, nir_var_mem_ssbo,
-            nvk_buffer_addr_format(rs->storage_buffers));
+            nvk_ssbo_addr_format(pdev, rs));
    NIR_PASS(_, nir, nir_lower_explicit_io, nir_var_mem_ubo,
-            nvk_buffer_addr_format(rs->uniform_buffers));
+            nvk_ubo_addr_format(pdev, rs));
    NIR_PASS(_, nir, nir_shader_intrinsics_pass,
-            lower_load_global_constant_offset_instr, nir_metadata_none, NULL);
+            lower_load_intrinsic, nir_metadata_none, NULL);
 
    if (!nir->info.shared_memory_explicit_layout) {
       NIR_PASS(_, nir, nir_lower_vars_to_explicit_types,
@@ -891,6 +922,13 @@ nvk_shader_get_executable_statistics(
 
    assert(executable_index == 0);
 
+   vk_outarray_append_typed(VkPipelineExecutableStatisticKHR, &out, stat) {
+      WRITE_STR(stat->name, "Instruction count");
+      WRITE_STR(stat->description, "Number of instructions used by this shader");
+      stat->format = VK_PIPELINE_EXECUTABLE_STATISTIC_FORMAT_UINT64_KHR;
+      stat->value.u64 = shader->info.num_instrs;
+   }
+
    vk_outarray_append_typed(VkPipelineExecutableStatisticKHR, &out, stat) {
       WRITE_STR(stat->name, "Code Size");
       WRITE_STR(stat->description,
